/*
 * libtaskstats - Library for easy access to taskstats interface
 *
 * Copyright IBM Corp. 2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 */

#ifndef _LIBTASKSTATS_H_
#define _LIBTASKSTATS_H_

#include <sys/time.h>
#include "taskstats_new.h"

#ifdef __cplusplus
extern "C" {
#endif

#define TS_VERSION 1 // XXX

#define TS_ERR_VERSION	-1

struct ts_socket;

extern int __ts_init(unsigned int version_compile,
		     unsigned int taskstats_size);

static inline int ts_init(void)
{
	return __ts_init(TASKSTATS_VERSION, sizeof(struct taskstats));
}

/**
 * ts_version - Return kernel taskstats version
 */
extern int ts_version(void);

/**
 * ts_cmd_get_pid - Read taskstats for given PID
 * @tsh: Taskstats handle
 * @pid: Process ID
 * @ts:  Pointer to taskstats buffer
 *
 * Returns 0 on success and negative error code otherwise.
 */
extern int ts_cmd_get_pid(struct ts_socket *tsh, __u32 pid,
			  struct taskstats *ts);

/**
 * ts_cmd_get_pids - Read taskstats vector for PIDs greater or equal
 * @tsh:    Taskstats handle
 * @pid:    Start process ID
 * @cnt:    Maximum number of taskstats
 * @time:   Timestamp, read only taskstats for tasks that have been or still
 *          are in state TASK_RUNNING between the given 'time' and now
 *          (correlates to taskstats.time_ns).
 * @ts_vec: taskstats vector (must have cnt elements)
 *
 * Read at most 'cnt' taskstats beginning with 'pid' that have been running
 * after time 'time'.
 *
 * Returns number of copied taskstats on success and a negative error code
 * otherwise. If the returned number is smaller than the requested number, no
 * more tasks are available.
 *
 * Examples:
 *
 * Tasks 5,7,9,11 and 20 are running:
 * ts_cmd_get_pids(tsh, 5, 3, 0, ts_vec) will return taskstats for PIDs 7,9
 * and 11.
 *
 * Tasks 5,7,11 and 20 are running, task 9 is sleeping and was last scheduled
 * at "last_task_9_sched_time":
 * ts_cmd_get_pids(tsh, 5, 3, last_task_9_sched_time, ts_vec) will return
 * taskstats for PIDs 7,11 and 20.
 */
extern int ts_cmd_get_pids(struct ts_socket *tsh, __u32 pid, int cnt,
			   __u64 time, __u64 *time_new,
			   struct taskstats *ts_vec);

extern int ts_cmd_get_tgids(struct ts_socket *tsh, __u32 tgid, int cnt,
			    __u64 time, struct taskstats *ts_vec);

extern int ts_read(struct ts_socket *tsh, struct taskstats *ts);

extern int ts_exit_register(struct ts_socket *tsh, const char *cpumask);
extern int ts_exit_deregister(struct ts_socket *tsh, const char *cpumask);
extern struct ts_socket *ts_socket_create(void);
extern void ts_socket_close(struct ts_socket *tsh);
extern int ts_socket_get_fd(struct ts_socket *tsh);

extern int ts_snapshot_register(struct ts_socket *tsh, const char *dir);

#ifdef __cplusplus
}
#endif

#endif
