/*
 * libtaskstats_snap - Snapshot library for easy access to taskstats interface
 *
 * Copyright IBM Corp. 2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "libtaskstats_snap.h"
#include "helper.h"

static struct {
	int			snap_initialized;
	struct ts_socket	*ts_socket;
} l;

struct ts_snap_handle {
	__u64		update_ns;
	int		prev_snap_cnt;
	int		flags;
};

struct ts_snap {
	__u64			update_ns;

	struct taskstats	**ts_sleep_vec;
	int			ts_sleep_cnt;

	struct taskstats	**ts_run_vec;
	int			ts_run_cnt;

	struct taskstats	**ts_exit_vec;
	int			ts_exit_cnt;

	struct taskstats	**ts_fork_vec;
	int			ts_fork_cnt;

	struct ts_snap_handle	*tss_h;
};

struct ts_snap_iter {
	struct taskstats	**ts_vec;
	int			curr;
	int			cnt;
};

__u64 ts_snap_update_ns(struct ts_snap *tss)
{
	return tss->update_ns;
}

#if 0
struct ts_snap_handle *ts_snap_init(int flags)
{
	struct ts_snap_handle *tss_h;

	tss_h = malloc(sizeof(*tss_h));
	if (tss_h) {
		tss_h->pref_snap_cnt = -1;
		tss_h->flags = flags;
	}
	return tss_h;
}
#endif

static void ts_run_read(struct ts_snap *tss)
{
	struct taskstats *ts_vec = NULL;
	__u32 pid = 0;
	__u64 time_ns;
	int rc, i;
	static int req_cnt = 0; // XXX reduce again later

	do {
		req_cnt += 100;
		ts_vec = realloc(ts_vec, sizeof(struct taskstats) * req_cnt);
		if (!ts_vec)
			ERR_EXIT("Malloc failed\n");
		rc = ts_cmd_get_pids(l.ts_socket, pid, req_cnt, 0,
				     &time_ns, ts_vec);
		if (rc < 0)
			ERR_EXIT("ts_cmd_get_pids failed\n");
	} while (rc == req_cnt);

	tss->ts_run_vec = realloc(tss->ts_run_vec, sizeof(void *) *
				  (tss->ts_run_cnt + rc));
	for (i = 0; i < rc; i++) {
		tss->ts_run_vec[i + tss->ts_run_cnt] =
			malloc(sizeof(ts_vec[i]));
		memcpy(tss->ts_run_vec[i + tss->ts_run_cnt],
		       &ts_vec[i], sizeof(ts_vec[i]));
	}
	tss->update_ns = time_ns;
	tss->ts_run_cnt += rc;
	free(ts_vec);
}

#if 0
static void ts_exit_read(struct ts_snap *tss)
{
	ERR_EXIT("ts_exit_read not implemented\n");
}

static void read_snap(struct ts_snap *tss)
{
	if (tss->tss_h->flags & TS_SNAP_TASK_EXIT)
		ts_exit_read(tss);
	if (tss->tss_h->flags & TS_SNAP_TASK_RUN)
		ts_run_read(tss);
}
#endif

void l_snap_init(void)
{
	if (l.snap_initialized)
		return;
	ts_init();
	l.ts_socket = ts_socket_create();
	if (!l.ts_socket)
		perror("tssocket failed\n");
	l.snap_initialized = 1;
}

struct ts_snap *l_ts_snap_alloc(struct ts_snap_handle *tss_h)
{
	struct ts_snap *tss;

	tss = calloc(1, sizeof(*tss));
	if (tss)
		tss->tss_h = tss_h;
	return tss;
}

void ts_snap_free(struct ts_snap *tss)
{
	int i;

	for (i = 0; i < tss->ts_run_cnt; i++)
		free(tss->ts_run_vec[i]);
	free(tss->ts_run_vec);
	for (i = 0; i < tss->ts_exit_cnt; i++)
		free(tss->ts_exit_vec[i]);
	free(tss->ts_exit_vec);
	free(tss);
}

struct ts_snap_handle *ts_snap_init(int flags)
{
	struct ts_snap_handle *tss_h;
	struct ts_snap *first_snap;

	tss_h = malloc(sizeof(*tss_h));
	if (!tss_h)
		return NULL;
	l_snap_init();
	tss_h->flags = flags;
	first_snap = l_ts_snap_alloc(tss_h);
	ts_run_read(first_snap);
	tss_h->prev_snap_cnt = first_snap->ts_run_cnt;
	tss_h->update_ns = 0;
	ts_snap_free(first_snap);
	return tss_h;
}

struct ts_snap *ts_snap_next(struct ts_snap_handle *tss_h)
{
	struct ts_snap *tss;

	tss = l_ts_snap_alloc(tss_h);
	if (!tss)
		return NULL;
	ts_run_read(tss);
	tss_h->prev_snap_cnt = tss->ts_run_cnt;
	tss_h->update_ns = tss->update_ns;
	return tss;
}

struct ts_snap_iter *ts_snap_iter_new(struct ts_snap *tss, int flags)
{
	struct ts_snap_iter *tss_i;

	tss_i = malloc(sizeof(*tss_i));
	tss_i->curr = 0;
	switch (flags) {
	case TS_SNAP_TASK_EXIT:
		tss_i->cnt = tss->ts_exit_cnt;
		tss_i->ts_vec = tss->ts_exit_vec;
		break;
	case TS_SNAP_TASK_RUN:
		tss_i->cnt = tss->ts_run_cnt;
		tss_i->ts_vec = tss->ts_run_vec;
		break;
	default:
		free(tss_i);
		return NULL;
	}
	return tss_i;
}

struct taskstats *ts_snap_iter_next(struct ts_snap_iter *tss_i)
{
	if (tss_i->cnt == tss_i->curr)
		return NULL;
	tss_i->curr++;
	return tss_i->ts_vec[tss_i->curr - 1];
}

void ts_snap_iter_free(struct ts_snap_iter *tss_i)
{
	free(tss_i);
}

unsigned long ts_snap_iter_cnt(struct ts_snap_iter *tss_i)
{
	return tss_i->cnt;
}

void ts_snap_exit(struct ts_snap_handle *tss_h)
{
	// XXX
	return;
}
