/*
 * ptop - Precise top command using the taskstats interface
 *
 * Copyright IBM Corp. 2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 */

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <ctype.h>
#include "ptop.h"
#include "sd.h"
#include "helper.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <poll.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <signal.h>
#include "libtaskstats.h"
#include "libtaskstats_snap.h"

static struct {
	struct ts_socket	*sd;
	u64			update_time_us;
	u64			last_update_ns;
	u64			new_update_ns;
	struct ts_snap_handle	*tss_h;
} l;

u64 start_time_us;
u64 end_time_us;
u64 total_cpu_sys_us;
u64 total_cpu_user_us;
u64 total_cpu_steal_us;
u64 total_update_cnt;

void print_cpu_times(void)
{
	printf("Total system time: %llu\n", total_cpu_sys_us);
	printf("Total user time..: %llu\n", total_cpu_user_us);
	printf("Total steal time.: %llu\n", total_cpu_steal_us);
	printf("-------------------------------------------------------\n");
	printf("Total update cnt.: %llu\n", total_update_cnt);
	printf("Total CPU time...: %llu\n",
	       total_cpu_sys_us + total_cpu_user_us + total_cpu_steal_us);
	printf("Total time.......: %llu\n", end_time_us - start_time_us);
}

/*
 * Compensate for task running over update timestamp
 */ 
static void l_adjust_times(struct taskstats *ts_new, struct taskstats *ts_old)
{
	long double factor;

	if (!g.initialized)
		return;
	if (l.new_update_ns >= ts_new->time_ns)
		return;

	factor = (long double) (ts_new->time_ns - l.last_update_ns) /
		(long double) (l.new_update_ns - l.last_update_ns) - 1;
	ts_new->ac_utime -= (ts_new->ac_utime - ts_old->ac_utime) * factor;
	ts_new->ac_stime -= (ts_new->ac_stime - ts_old->ac_stime) * factor;
	ts_new->ac_sttime -= (ts_new->ac_sttime - ts_old->ac_sttime) * factor;
	ts_new->ac_cutime -= (ts_new->ac_cutime - ts_old->ac_cutime) * factor;
	ts_new->ac_cstime -= (ts_new->ac_cstime - ts_old->ac_cstime) * factor;
	ts_new->ac_csttime -= (ts_new->ac_csttime - ts_old->ac_csttime) *
		factor;
}

static void l_calc_diff(u64 *diff_us, u64 *total_us, u64 old_us, u64 new_us,
			long double interval_us)
{
	long double __diff_us;

	__diff_us = new_us - old_us;
	if (total_us)
		*total_us += (u64) (__diff_us + 0.5);
	__diff_us /= (interval_us / 1000000);
	*diff_us = (u64) (__diff_us + 0.5);
	total_update_cnt++;
}

#ifndef __s390x__
/*
 * Calculate steal time for architectures that do not save the
 * steal time correctly in taskstats.
 */
static u64 l_calc_sttime_old(struct taskstats *ts_new,
			     struct taskstats *ts_old,
			     long double interval_us)
{
#if 0
	long double cpu_steal_diff;

	cpu_steal_diff =
		((long double)
		 (ts_new->cpu_run_virtual_total -
		  ts_old->cpu_run_virtual_total))
		 -
		 (ts_new->cpu_run_real_total - ts_old->cpu_run_real_total);
	if (cpu_steal_diff < 0)
		cpu_steal_diff = 0;
	if (g.initialized)
		total_cpu_steal_us += (u64) cpu_steal_diff / 1000;
	cpu_steal_diff /= (interval_us / 1000);
	return (u64) cpu_steal_diff;
#else
	return 0;
#endif
}
#endif

static void l_taskstats_diff(struct sd_task *proc, struct taskstats *ts_new)
{
	struct taskstats *ts_old = &proc->ts;
	long double interval_us;

	l_adjust_times(ts_new, ts_old);

	interval_us =
		((long double) (l.new_update_ns - l.last_update_ns)) / 1000;

	/* CPU user */
	l_calc_diff(&proc->cpu_utime_diff, &total_cpu_user_us, ts_old->ac_utime,
		    ts_new->ac_utime, interval_us);

	/* CPU system */
	l_calc_diff(&proc->cpu_stime_diff, &total_cpu_sys_us, ts_old->ac_stime,
		    ts_new->ac_stime, interval_us);

#ifdef __s390x__
	/* CPU steal */
	l_calc_diff(&proc->cpu_sttime_diff, &total_cpu_steal_us,
		    ts_old->ac_sttime, ts_new->ac_sttime, interval_us);
#else
	proc->cpu_sttime_diff = l_calc_sttime_old(ts_new, ts_old, interval_us);
#endif

	/* CPU delay */
	l_calc_diff(&proc->delay_cpu_diff, NULL,
		    ts_old->cpu_delay_total / 1000,
		    ts_new->cpu_delay_total / 1000, interval_us);

	/* Swap delay */
	l_calc_diff(&proc->delay_swap_diff, NULL,
		    ts_old->swapin_delay_total / 1000,
		    ts_new->swapin_delay_total / 1000, interval_us);

	/* BLK io delay */
	l_calc_diff(&proc->delay_blkio_diff, NULL,
		    ts_old->blkio_delay_total / 1000,
		    ts_new->blkio_delay_total / 1000, interval_us);

	/* Memory delay */
	l_calc_diff(&proc->delay_mem_diff, NULL,
		    ts_old->freepages_delay_total / 1000,
		    ts_new->freepages_delay_total / 1000, interval_us);

	/* Total diff */
#if 1
	proc->total_diff = proc->cpu_utime_diff + proc->cpu_stime_diff +
		proc->cpu_sttime_diff;
	// + ts_old->cpu_delay_total_diff;
#else
	proc->total_diff = proc->cpu_utime_diff + proc->cpu_stime_diff +
		ts_old->cpu_delay_total_diff + proc->cpu_sttime_diff +
		ts_old->blkio_delay_total_diff +
		ts_old->swapin_delay_total_diff;
#endif

#ifdef WITH_CTIME
	/*
	 * Child CPU times
	 */
	proc->cpu_cutime_diff = ts_new->ac_cutime - ts_old->ac_cutime;
	proc->cpu_cstime_diff = ts_new->ac_cstime - ts_old->ac_cstime;
	proc->cpu_csttime_diff = ts_new->ac_csttime - ts_old->ac_csttime;
#endif
	/* Mem core */
	if (!ts_new->ac_utime || !ts_new->ac_stime)
		return;

	proc->mem_core_diff_kib = ts_new->coremem * 1024 -
		proc->mem_core_kib;
	proc->mem_core_diff_kib /= (ts_new->ac_utime + ts_new->ac_stime);
	proc->mem_core_kib = ts_new->coremem * 1024 / ts_new->ac_stime;
	/* Mem virt */
	proc->mem_virt_kib = ts_new->virtmem * 1024 / ts_new->ac_stime;
}

static void l_taskstats_add(struct sd_task *proc, struct taskstats *ts_new)
{
	if (!start_time_us)
		start_time_us = ts_new->time_ns / 1000;
	end_time_us = ts_new->time_ns / 1000;

	if (g.initialized)
		l_taskstats_diff(proc, ts_new);

	/* Parent PID */
#if 0
	if (proc->ts.ac_ppid == 0 && ts_new->ac_pid == 0
	    && ts_new->ac_pid != 1 && ts_new->ac_pid != 2)
		ERR_EXIT("PPID=0, pid=%u com=%s\n", ts_new->ac_pid,
			ts_new->ac_comm); // XXX
#endif
	memcpy(&proc->ts, ts_new, sizeof(*ts_new));
	strcpy(sd_task_item_name_process(proc), ts_new->ac_comm);
}


#ifdef WITH_SNAP

static void l_ts_snap_init(void)
{
	// l.tss_h = ts_snap_init(TS_SNAP_TASK_EXIT|TS_SNAP_TASK_RUN);
	l.tss_h = ts_snap_init(TS_SNAP_TASK_RUN);
	if (!l.tss_h)
		ERR_EXIT("snap init failed\n");
}

/*
 * Fill System Data
 */
static void l_sd_task_root_fill(struct sd_task *task)
{
	struct ts_snap_iter *tss_i;
	struct sd_task *proc;
	struct taskstats *ts;
	struct ts_snap *tss;
	char pid_str[100];

	tss = ts_snap_next(l.tss_h);
	tss_i = ts_snap_iter_new(tss, TS_SNAP_TASK_RUN);
	if (!tss_i)
		ERR_EXIT("ts_snap_iter_new failed\n");

	l.new_update_ns = ts_snap_update_ns(tss);
	l.update_time_us = l.new_update_ns / 1000;
	do {
		ts = ts_snap_iter_next(tss_i);
		if (!ts)
			break;
		sprintf(pid_str, "%d", ts->ac_pid);
		proc = sd_task_get(task, pid_str);
		if (!proc)
			proc = sd_task_new(task, pid_str);
		l_taskstats_add(proc, ts);
		proc->i.calculated = 2; // XXX
		sd_task_commit(proc);
	} while (1);
	ts_snap_iter_free(tss_i);
#if 0
	tss_i = ts_snap_iter_new(tss, TS_SNAP_TASK_EXIT);
	if (!tss_i)
		ERR_EXIT("ts_snap_iter_new failed\n");
	do {
		ts = ts_snap_iter_next(tss_i);
		if (!ts)
			break;
		// printf("XXX: %i\n", ts_new->ac_pid);
		sprintf(pid_str, "%d", ts_new->ac_pid);
		proc = sd_task_get(task, pid_str);
		if (!proc)
			proc = sd_task_new(task, pid_str);
		l_taskstats_add(proc, ts);
		sd_task_commit(proc);
		proc->i.calculated = 2; // XXX
	} while (1);
	ts_snap_iter_free(tss_i);
#endif

	ts_snap_free(tss);
	l.last_update_ns = l.new_update_ns;
	sd_task_commit(task);
}

#else

#ifdef WITH_CMD_PIDS

#define PIDS_PER_QUERY	50

/*
 * Fill System Data
 */
static void l_sd_task_root_fill(struct sd_task *task)
{
	struct taskstats taskstats_vec[PIDS_PER_QUERY];
	struct sd_task *proc;
	int pid = 0, i, rc;
	char pid_str[100];

	do {
		rc = ts_cmd_get_pids(l.sd, pid, PIDS_PER_QUERY,
				     l.last_update_ns, &l.new_update_ns,
				     &taskstats_vec[0]);
		for (i = 0; i < rc; i++) {
			sprintf(pid_str, "%d", taskstats_vec[i].ac_pid);
			proc = sd_task_get(task, pid_str);
			if (!proc)
				proc = sd_task_new(task, pid_str);
			l_taskstats_add(proc, &taskstats_vec[i]);
			sd_task_commit(proc);
		}
		if (rc < PIDS_PER_QUERY)
			break;
		pid = taskstats_vec[rc - 1].ac_pid + 1;
	} while (1);
	l.last_update_ns = l.new_update_ns;
	sd_task_commit(task);
}

#else /* WITH_CMD_PIDS */

static void l_read_proc(struct sd_task *proc)
{
	struct taskstats taskstats;
	__u32 pid = atoi(proc->id);

	memset(&taskstats, 0, sizeof(taskstats));
	ts_cmd_get_pid(l.sd, pid, &taskstats);
	// printf("add: %s %llu\n", taskstats.ac_comm, taskstats.ac_utime);
	l_taskstats_add(proc, &taskstats);
	sd_task_commit(proc);
}

/*
 * Fill System Data
 */
static void l_sd_task_root_fill(struct sd_task *task)
{
	struct dirent *direntp;
	struct sd_task *guest;
	struct timeval tv;
	DIR *fh_dir;

	fh_dir = opendir("/proc/");

	while ((direntp = readdir(fh_dir))) {
		if (!isdigit(direntp->d_name[0]))
			continue;
		guest = sd_task_get(task, direntp->d_name);
		if (!guest)
			guest = sd_task_new(task, direntp->d_name);
		gettimeofday(&tv, NULL);
		l.new_update_ns = tv.tv_sec * 1000000 + tv.tv_usec;
		l.new_update_ns *= 1000;
		l_read_proc(guest);
	}
	l.last_update_ns = l.new_update_ns;
	sd_task_commit(task);
	closedir(fh_dir);
}

#endif /* WITH_CMD_PIDS */
#endif /* WITH_SNAP */

/*
 * Update tasktem data
 */
static void l_sd_update(void)
{
	struct sd_task *root = sd_task_root_get();

	sd_task_update_start(root);
	l_sd_task_root_fill(root);
	sd_task_update_end(root, l.update_time_us);
}

/*
 * Supported tasktem items
 */
static struct sd_task_item *l_task_item_vec[] = {
	&sd_task_item_tgid,
	&sd_task_item_ppid,
	&sd_task_item_cpu_utime_diff,
	&sd_task_item_cpu_stime_diff,
	&sd_task_item_cpu_sttime_diff,
#ifdef WITH_CTIME
	&sd_task_item_cpu_cutime_diff,
	&sd_task_item_cpu_cstime_diff,
	&sd_task_item_cpu_csttime_diff,
	&sd_task_item_cpu_cutime,
	&sd_task_item_cpu_cstime,
	&sd_task_item_cpu_csttime,
#endif
	&sd_task_item_cpu_user,
	&sd_task_item_cpu_sys,
	&sd_task_item_cpu_steal,
	&sd_task_item_delay_cpu,
	&sd_task_item_delay_cpu_diff,
	&sd_task_item_delay_swap_diff,
	&sd_task_item_delay_blkio_diff,
	&sd_task_item_delay_mem_diff,
	&sd_task_item_total_diff,
	&sd_task_item_etime,
	&sd_task_item_page_fault_min,
	&sd_task_item_page_fault_maj,
	&sd_task_item_csw_vol,
	&sd_task_item_csw_nonvol,
	&sd_task_item_mem_core,
	&sd_task_item_mem_virt,
	&sd_task_item_name,
	NULL,
};

/*
 * Default tasktem items
 */
static struct sd_task_item *l_task_item_enable_vec[] = {
	&sd_task_item_cpu_utime_diff,
	&sd_task_item_cpu_stime_diff,
	&sd_task_item_cpu_sttime_diff,
#ifdef WITH_CTIME
	&sd_task_item_cpu_cutime_diff,
	&sd_task_item_cpu_cstime_diff,
	&sd_task_item_cpu_csttime_diff,
	// &sd_task_item_cpu_cutime,
	// &sd_task_item_cpu_cstime,
	// &sd_task_item_cpu_csttime,
#endif
	&sd_task_item_delay_cpu_diff,
	// &sd_task_item_delay_blkio_diff,
	// &sd_task_item_delay_swap_diff,
	// &sd_task_item_delay_mem_diff,
	&sd_task_item_total_diff,
	&sd_task_item_etime,
	&sd_task_item_name,
	NULL,
};

/*
 * Define data gatherer structure
 */
static struct sd_dg dg_taskstats = {
	.update_task		= l_sd_update,
	.task_item_vec		= l_task_item_vec,
	.task_item_enable_vec	= l_task_item_enable_vec,
};


/*
 * Initialize z/VM debugfs data gatherer
 */
int dg_taskstats_init(void)
{
	if (ts_init())
		ERR_EXIT("ts_init failed\n");
#ifdef WITH_SNAP
	l_ts_snap_init();
#else
#ifdef WITH_CMD_PIDS
	if (ts_version() < 8)
		ERR_EXIT("Invalid kernel taskstats version %d (required 8)\n",
			 ts_version());
#else
	if (ts_version() < 5)
		ERR_EXIT("Invalid kernel taskstats version %d (required 5)\n",
			 ts_version());
#endif
#endif
	if ((l.sd = ts_socket_create()) == NULL)
		ERR_EXIT("error creating Netlink socket\n");

	sd_dg_register(&dg_taskstats);
	return 0;
}
