/*
 * ptop - Precise top command using the taskstats interface
 *
 * Command line options, window definition, print functions, etc.
 *
 * Copyright IBM Corp. 2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 */

#ifndef PTOP_H
#define PTOP_H

#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>
#include <termios.h>
#include "list.h"
#include "helper.h"
#include "table.h"
#include "nav_desc.h"

#define PTOP_OPT_DEFAULT_DELAY	2
#define PTOP_MAX_WIN_DEPTH		4
#define PTOP_MAX_LINE			512
#define PROG_NAME			"ptop"

/*
 * Options info
 */
struct ptop_str_vec_opt {
	unsigned int		specified;
	char			**vec;
	unsigned int		cnt;
};

struct ptop_col_vec_opt {
	unsigned int		specified;
	struct table_col_spec	**vec;
	unsigned int		cnt;
};

struct ptop_win_opts {
	struct ptop_str_vec_opt	task;
	struct ptop_col_vec_opt	fields;
	unsigned int			sort_field_specified;
	char				sort_field;
};

struct ptop_opts {
	unsigned int			win_specified;
	unsigned int			batch_mode_specified;
	unsigned int			iterations_specified;
	unsigned int			iterations;
	unsigned int			iterations_act;

	struct ptop_win		*cur_win;
	struct ptop_str_vec_opt	cpu_types;

	int				delay_s;
	int				delay_us;
};

/*
 * Curses info
 */
struct ptop_curses {
	int			row_cnt;
	int			col_cnt;
	char			line[PTOP_MAX_LINE];
	int			x;
	int			y;
	int			initialized;
};

/*
 * Window info
 */
struct ptop_win_info {
	struct ptop_win	*cur;
	struct ptop_win	*prev[PTOP_MAX_WIN_DEPTH];
	unsigned int		prev_cnt;
};

/*
 * Globals definition
 */
struct ptop_globals {
	struct ptop_opts	o;
	struct ptop_curses	c;
	struct ptop_win_info	w;
	const char		*prog_name;
	struct ptop_win	*win_cpu_types;
	int			initialized;
};

extern struct ptop_globals g;

/*
 * Print functions
 */
#define ptop_printf_pos(y, x, p...) \
	do { \
		if (g.o.batch_mode_specified) \
			printf(p); \
		else { \
			int len; \
			len = snprintf(g.c.line, sizeof(g.c.line) - 1, p); \
			len = MIN(len, (g.c.col_cnt - (x))); \
			if (len > 0) { \
				mvaddnstr((y), (x), g.c.line, len); \
			} \
		} \
	} while (0)

#define ptop_printf(p...) \
	do { \
		if (g.o.batch_mode_specified) \
			printf(p); \
		else { \
			int len; \
			len = snprintf(g.c.line, sizeof(g.c.line) - 1, p); \
			len = MIN(len, (g.c.col_cnt - g.c.x)); \
			if (len > 0) { \
				mvaddnstr(g.c.y, g.c.x, g.c.line, len); \
				g.c.x += len; \
			} \
		} \
	} while (0)

static inline void ptop_printf_init(void)
{
	g.c.x = 0;
	g.c.y = 0;
}

static inline void ptop_print_seek_back(int i)
{
	unsigned int cnt = MAX(g.c.col_cnt - g.c.x - i, 0);

	if (g.o.batch_mode_specified)
		return;
	if (cnt) {
		memset(g.c.line, ' ', cnt);
		assert(cnt < sizeof(g.c.line));
		g.c.line[cnt] = 0;
		addstr(g.c.line);
	}
	g.c.x = g.c.col_cnt - i;
}

static inline void ptop_print_nl(void)
{
	unsigned int cnt = MAX(g.c.col_cnt - g.c.x, 0);

	if (g.o.batch_mode_specified) {
		printf("\n");
		return;
	}
	if (cnt) {
		memset(g.c.line, ' ', g.c.col_cnt - g.c.x);
		assert(cnt < sizeof(g.c.line));
		g.c.line[cnt] = 0;
		addstr(g.c.line);
	}
	g.c.x = 0;
	g.c.y++;
}

/*
 * ptop windows
 */

enum ptop_win_action {
	WIN_SWITCH,
	WIN_KEEP,
};

extern enum ptop_win_action ptop_process_input_timeout(void);
extern enum ptop_win_action ptop_process_input(void);
extern enum ptop_win_action win_switch(struct ptop_win *w);
extern enum ptop_win_action win_back(void);

struct ptop_win;
struct ptop_win {
	enum ptop_win_action	(*process_input)(struct ptop_win *w, int c);
	void			(*update_term)(struct ptop_win *w);
	void			(*run)(struct ptop_win *w);
	const char		*id;
	const char		*desc;
	struct nav_desc		**desc_normal_vec;
	struct nav_desc		**desc_select_vec;
	struct nav_desc		**desc_general_vec;
	struct ptop_win_opts	opts;
};

/*
 * Window task_list
 */
extern struct ptop_win win_task_list;
extern void win_task_list_init(void);

/*
 * Window task
 */
extern struct ptop_win win_task;
extern void win_task_set(const char *task_id);
extern void win_task_init(void);

/*
 * Window cpu_types
 */
extern void win_cpu_types_init(void);

/*
 * Misc functions
 */
extern void ptop_update_term(void);
extern void ptop_exit(int rc);
extern void ptop_text_mode(void);

extern int dg_taskstats_init(void);
#endif /* PTOP_H */
