/*
 * ptop - Precise top command using the taskstats interface
 *
 * System data module: Provide database for task data (e.g. CPU and memory)
 *
 * Copyright IBM Corp. 2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 */

#ifndef SD_H
#define SD_H

#include "helper.h"
#include "table.h"
#include "taskstats_new.h"

#define SD_DG_INIT_INTERVAL_MS	200
#define SD_TASK_ID_SIZE		9

/*
 * System Name
 */
struct sd_task_name {
	char	process[100];
};

struct sd_task;

/*
 * SD info
 */
struct sd_info {
	u8		calculated;
	u8		active;
	struct sd_task	*parent;
};


/*
 * SD System (can be e.g. CEC, VM or guest/LPAR)
 */
struct sd_task {
	struct list		list;
	struct sd_info		i;
	u64			update_time_us;
	u32			child_cnt;
	u32			child_cnt_active;
	struct list		child_list;
	char			id[SD_TASK_ID_SIZE];
	struct sd_task_name	name;
	u64			time_ns;
	struct taskstats	ts;
	u64			cpu_utime_diff;
	u64			cpu_stime_diff;
	u64			cpu_sttime_diff;
	u64			cpu_cutime_diff;
	u64			cpu_cstime_diff;
	u64			cpu_csttime_diff;
	u64			delay_cpu_diff;
	u64			delay_swap_diff;
	u64			delay_blkio_diff;
	u64			delay_mem_diff;
	u64			total_diff;
	u64			mem_core_kib;
	u64			mem_core_diff_kib;
	u64			mem_virt_kib;
};

#define sd_task_id(task) ((task)->id)
#define sd_task_name_os(task) ((task)->name.os)

void sd_task_update_start(struct sd_task *task);
void sd_task_update_end(struct sd_task *task, u64 update_time_us);
struct sd_task *sd_task_root_get(void);
struct sd_task *sd_task_get(struct sd_task *parent, const char *id);
struct sd_task *sd_task_new(struct sd_task *parent, const char *id);

static inline int sd_task_active(struct sd_task *task)
{
	return task->i.active;
}

static inline void sd_task_commit(struct sd_task *task)
{
	struct sd_task *parent = task->i.parent;

	task->i.calculated += 1;  // XXX
	task->i.active = 1;
	if (parent)
		parent->child_cnt_active++;
}

/*
 * Item types
 */
enum sd_item_type {
	SD_TYPE_U16,
	SD_TYPE_U32,
	SD_TYPE_U64,
	SD_TYPE_S64,
	SD_TYPE_STR,
};

/*
 * Task item
 */
struct sd_task_item {
	struct table_col	table_col;
	enum sd_item_type	type;
	int			offset;
	char			*desc;
	int			info;
	u64 (*fn_u64)(struct sd_task_item *, struct sd_task *);
	s64 (*fn_s64)(struct sd_task_item *, struct sd_task *);
};

#define sd_task_item_table_col(item) (&item->table_col)
#define sd_task_item_type(item) (item->type)
#define sd_task_item_name_process(item) (item->name.process)

extern int sd_task_item_available(struct sd_task_item *item);
extern int sd_task_item_cnt(void);

/*
 * Item access functions
 */
static inline u64 sd_task_item_u64(struct sd_task *task,
				  struct sd_task_item *item)
{
	return item->fn_u64(item, task);
}

static inline s64 sd_task_item_s64(struct sd_task *task,
				  struct sd_task_item *item)
{
	return item->fn_s64(item, task);
}

static inline char *sd_task_item_str(struct sd_task *task,
				    struct sd_task_item *item)
{
	return ((char *) task) + item->offset;
}

/*
 * Predefined System items
 */
extern struct sd_task_item sd_task_item_cpu_user;
extern struct sd_task_item sd_task_item_cpu_sys;
extern struct sd_task_item sd_task_item_cpu_steal;

extern struct sd_task_item sd_task_item_cpu_utime_diff;
extern struct sd_task_item sd_task_item_cpu_stime_diff;
extern struct sd_task_item sd_task_item_cpu_sttime_diff;

extern struct sd_task_item sd_task_item_cpu_cutime;
extern struct sd_task_item sd_task_item_cpu_cstime;
extern struct sd_task_item sd_task_item_cpu_csttime;

extern struct sd_task_item sd_task_item_cpu_cutime_diff;
extern struct sd_task_item sd_task_item_cpu_cstime_diff;
extern struct sd_task_item sd_task_item_cpu_csttime_diff;

extern struct sd_task_item sd_task_item_delay_cpu;
extern struct sd_task_item sd_task_item_delay_cpu_diff;
extern struct sd_task_item sd_task_item_delay_swap_diff;
extern struct sd_task_item sd_task_item_delay_blkio_diff;
extern struct sd_task_item sd_task_item_delay_mem_diff;
extern struct sd_task_item sd_task_item_total_diff;
extern struct sd_task_item sd_task_item_etime;
extern struct sd_task_item sd_task_item_page_fault_min;
extern struct sd_task_item sd_task_item_page_fault_maj;
extern struct sd_task_item sd_task_item_csw_vol;
extern struct sd_task_item sd_task_item_csw_nonvol;
extern struct sd_task_item sd_task_item_name;
extern struct sd_task_item sd_task_item_mem_core;
extern struct sd_task_item sd_task_item_mem_virt;
extern struct sd_task_item sd_task_item_tgid;
extern struct sd_task_item sd_task_item_ppid;

/*
 * Data gatherer backend
 */
struct sd_dg {
	void 			(*update_task)(void);
	struct sd_task_item	**task_item_vec;
	struct sd_task_item	**task_item_enable_vec;
};

void sd_dg_register(struct sd_dg *);

/*
 * Iterators
 */
#define sd_task_iterate(parent, task) \
	list_iterate(task, &parent->child_list, list)

#define sd_task_item_iterate(ptr, i) \
	for (i = 0; (ptr = sd.dg->task_item_vec[i]); i++)

#define sd_task_item_enable_iterate(ptr, i) \
	for (i = 0; (ptr = sd.dg->task_item_enable_vec[i]); i++)

/*
 * Offset macros
 */
#define SD_TASK_OFFSET(x) \
	((unsigned long)(void *)&(((struct sd_task *) NULL)->x))

/*
 * Misc
 */
void sd_update(void);
extern void sd_init(void);

static inline u64 l_sub_64(u64 x, u64 y)
{
	return x < y ? 0 : x - y;
}

struct sd_globals {
	struct sd_dg	*dg;
};

extern struct sd_globals sd;

#endif /* SD_H */
