/*
 * ptop - Precise top command using the taskstats interface
 *
 * System data module: Provide backend independent database for taskt data
 *                     (e.g. for CPU and memory data)
 *
 * Copyright IBM Corp. 2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 */

#include <string.h>
#include <time.h>
#include "sd.h"
#include "ptop.h"
#include "helper.h"
#include "opts.h"

/*
 * Internal globals for taskitem data
 */
static int		l_task_item_cnt;
static struct sd_task	*l_root_task;

/*
 * External globals for tasktem data
 */
struct sd_globals sd;

/*
 * Get root tasktem
 */
struct sd_task *sd_task_root_get(void)
{
	return l_root_task;
}

/*
 * Update tasktem data using the data gatherer
 */
void sd_update(void)
{
	sd.dg->update_task();
}

/*
 * Register a data gatherer
 */
void sd_dg_register(struct sd_dg *dg)
{
	struct timespec ts = {0, SD_DG_INIT_INTERVAL_MS * 1000000};
	struct sd_task_item *task_item;
	unsigned int i;

	sd.dg = dg;

	sd_task_item_iterate(task_item, i)
		l_task_item_cnt++;

	sd_update();
	g.initialized = 1;
	nanosleep(&ts, NULL);
	sd_update();
}

/*
 * Get tasktem by ID
 */
struct sd_task *sd_task_get(struct sd_task *parent, const char* id)
{
	struct sd_task *task;

	list_iterate(task, &parent->child_list, list) {
		if (strcmp(task->id, id) == 0)
			return task;
	}
	return NULL;
}

/*
 * Allocate and initialize new tasktem
 */
struct sd_task *sd_task_new(struct sd_task *parent, const char *id)
{
	struct sd_task *task_new;

	task_new = ht_zalloc(sizeof(*task_new));
	strncpy(task_new->id, id, sizeof(task_new->id));
	list_init(&task_new->child_list);
	list_init(&task_new->list);

	if (parent) {
		task_new->i.parent = parent;
		parent->child_cnt++;
		list_add_end(&task_new->list, &parent->child_list);
	}
	return task_new;
}

/*
 * Free tasktem
 */
static void sd_task_free(struct sd_task *task)
{
	ht_free(task);
}

/*
 * Start update cycle for tasktem
 */
void sd_task_update_start(struct sd_task *task)
{
	struct sd_task *child;

	task->i.active = 0;
	task->child_cnt_active = 0;

	list_iterate(child, &task->child_list, list)
		sd_task_update_start(child);
}

#ifdef WITH_CTIME

static void l_update_parent(struct sd_task *parent, struct sd_task *task)
{
	if (parent->cpu_cutime_diff >=
	    task->ts.ac_utime + task->ts.ac_cutime) {
		parent->cpu_cutime_diff -= task->ts.ac_utime;
		parent->cpu_cutime_diff -= task->ts.ac_cutime;
	} else {
		/*
		ERR_EXIT("XXX user: pid=%s (%s) ppid=%u(%s): %llu + %llu / %llu \n",
			 task->id,
			 sd_task_item_name_process(task),
			 task->ts.ac_ppid,
			 sd_task_item_name_process(parent),
			 task->ts.ac_utime, task->cpu_cutime_diff, parent->cpu_cutime_diff);
			 */
		parent->cpu_cutime_diff = 0;
	}
	if (parent->cpu_cstime_diff >=
	    task->ts.ac_stime + task->ts.ac_cstime) {
		parent->cpu_cstime_diff -= task->ts.ac_stime;
		parent->cpu_cstime_diff -= task->ts.ac_cstime;
	} else {
		/*
		ERR_EXIT("XXX sys: pid=%s (%s) ppid=%u(%s): %llu + %llu / %llu \n",
			 task->id,
			 sd_task_item_name_process(task),
			 task->ts.ac_ppid,
			 sd_task_item_name_process(parent),
			 task->ts.ac_stime, task->cpu_cstime_diff, parent->cpu_cstime_diff);
			*/
		parent->cpu_cstime_diff = 0;
	}
	if (parent->cpu_csttime_diff >=
	    task->ts.ac_sttime + task->ts.ac_csttime) {
		parent->cpu_csttime_diff -= task->ts.ac_sttime;
		parent->cpu_csttime_diff -= task->ts.ac_csttime;
	} else {
		/*
		ERR_EXIT("XXX steal: pid=%s (%s) ppid=%u(%s): %llu + %llu / %llu \n",
			 task->id,
			 sd_task_item_name_process(task),
			 task->ts.ac_ppid,
			 sd_task_item_name_process(parent),
			 task->ts.ac_sttime, task->cpu_csttime_diff, parent->cpu_csttime_diff);
			 */
		parent->cpu_csttime_diff = 0;
	}
}

static void l_update_child_exit(struct sd_task *root, struct sd_task *task)
{
	struct sd_task *parent, *child = task;
	int level = 1;

	do {
		char ppid[20];

		if (child->ts.ac_ppid == 0) {
			ERR_EXIT("PID=%s (%s) PPID=0\n", child->id, sd_task_item_name_process(child));
			// printf("PPID=0\n");
			return; // XXX taskstats problem
		}
		sprintf(ppid, "%u", child->ts.ac_ppid);
		parent = sd_task_get(root, ppid);
		if (parent->i.active) {
			l_update_parent(parent, task);
			break;
		}
		level++;
		child = parent;
	} while (1);
}

extern u64 total_cpu_sys_us;
extern u64 total_cpu_user_us;
extern u64 total_cpu_steal_us; // XXX

/*
 * End update cycle for tasktem
 */
static void l_task_update_end(struct sd_task *task, long double interval_us)
{
	struct sd_task *child, *tmp;

//	if (task->child_cnt_active == task->child_cnt)
//		return;

	list_iterate_safe(child, &task->child_list, list, tmp)
		if (!child->i.active)
			l_update_child_exit(task, child);

	list_iterate_safe(child, &task->child_list, list, tmp) {
		if (!child->i.active) {
#if 0
			child->cpu_utime_diff = 0;
			child->cpu_stime_diff = 0;
			child->cpu_sttime_diff = 0;
			child->total_diff = 0;
			child->ts.cpu_delay_total_diff = 0;
			continue; // XXX
#endif

			/* child has not been updated, remove it */
			list_del(&child->list);
			sd_task_free(child);
			continue;
		}
		/* Recursively update child */
		l_task_update_end(child, interval_us);
	}
	list_iterate_safe(child, &task->child_list, list, tmp) {
		if (g.initialized) {
			total_cpu_user_us += child->cpu_cutime_diff; // XXX
			total_cpu_sys_us += child->cpu_cstime_diff;
			total_cpu_steal_us += child->cpu_csttime_diff;
		}
		child->cpu_cutime_diff =
			((long double) child->cpu_cutime_diff) /
			(interval_us / 1000000) + 0.5;
		child->cpu_cstime_diff =
			((long double) child->cpu_cstime_diff) /
			(interval_us / 1000000) + 0.5;
		child->cpu_csttime_diff =
			((long double) child->cpu_csttime_diff) /
			(interval_us / 1000000) + 0.5;
		child->total_diff += child->cpu_cutime_diff;
		child->total_diff += child->cpu_cstime_diff;
		child->total_diff += child->cpu_csttime_diff;
	}
	task->child_cnt = task->child_cnt_active;
}

/*
 * End update cycle for tasktem
 */
void sd_task_update_end(struct sd_task *task, u64 update_time_us)
{
	u64 interval_us = update_time_us - task->update_time_us;
	task->update_time_us = update_time_us;
	l_task_update_end(task, interval_us);
}

#else
static void l_task_update_end(struct sd_task *task)
{
	struct sd_task *child, *tmp;

	if (task->child_cnt_active == task->child_cnt)
		return;

	list_iterate_safe(child, &task->child_list, list, tmp) {
		if (!child->i.active) {
#if 0
			child->cpu_utime_diff = 0;
			child->cpu_stime_diff = 0;
			child->cpu_sttime_diff = 0;
			child->total_diff = 0;
			child->ts.cpu_delay_total_diff = 0;
			continue; // XXX
#endif
			/* child has not been updated, remove it */
			list_del(&child->list);
			sd_task_free(child);
			continue;
		}
		/* Recursively update child */
		l_task_update_end(child);
	}
	task->child_cnt = task->child_cnt_active;
}

/*
 * End update cycle for tasktem
 */
void sd_task_update_end(struct sd_task *task, u64 update_time_us)
{
	task->update_time_us = update_time_us;
	l_task_update_end(task);
}
#endif 

/*
 * Is tasktem item available?
 */
int sd_task_item_available(struct sd_task_item *item)
{
	struct sd_task_item *ptr;
	unsigned int i;

	sd_task_item_iterate(ptr, i) {
		if (item == ptr)
			return 1;
	}
	return 0;
}

/*
 * Number of tasktem items
 */
int sd_task_item_cnt(void)
{
	return l_task_item_cnt;
}

/*
 * Init tasktem data module
 */
void sd_init(void)
{
	l_root_task = sd_task_new(NULL, "root");
}
