/*
 * ptop - Precise top command using the taskstats interface
 *
 * Provide task items
 *
 * Copyright IBM Corp. 2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 */

#include "sd.h"

/*
 * Get u64 task item value from "task"
 */
static u64 l_task_item_u64(struct sd_task_item *item, struct sd_task *task)
{
	switch (item->type) {
	case SD_TYPE_U16:
		return *(u16 *)(((char *) task) + item->offset);
	case SD_TYPE_U32:
		return *(u32 *)(((char *) task) + item->offset);
	case SD_TYPE_U64:
		return *(u64 *)(((char *) task) + item->offset);
	case SD_TYPE_S64:
	case SD_TYPE_STR:
		break;
	}
	assert(0);
	return 0;
}

/*
 * Task item definitions
 */

struct sd_task_item sd_task_item_ppid = {
	.table_col = TABLE_COL_CNT_NONE('I', "ppid"),
	.offset = SD_TASK_OFFSET(ts.ac_ppid),
	.type	= SD_TYPE_U64,
	.desc	= "Parent PID",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_tgid = {
	.table_col = TABLE_COL_CNT_NONE('g', "tgid"),
	.offset = SD_TASK_OFFSET(ts.ac_tgid),
	.type	= SD_TYPE_U64,
	.desc	= "Thread group ID",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_utime_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 'u', "user"),
	.offset = SD_TASK_OFFSET(cpu_utime_diff),
	.type	= SD_TYPE_U64,
	.desc	= "CPU time (user) per second",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_stime_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 's', "sys"),
	.offset = SD_TASK_OFFSET(cpu_stime_diff),
	.type	= SD_TYPE_U64,
	.desc	= "CPU time (system) per second",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_sttime_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 't', "ste"),
	.offset = SD_TASK_OFFSET(cpu_sttime_diff),
	.type	= SD_TYPE_U64,
	.desc	= "CPU time (steal) per second",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_cutime_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 'r', "cuser"),
	.offset = SD_TASK_OFFSET(cpu_cutime_diff),
	.type	= SD_TYPE_U64,
	.desc	= "CPU time (user) per second of childs",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_cstime_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 'y', "csys"),
	.offset = SD_TASK_OFFSET(cpu_cstime_diff),
	.type	= SD_TYPE_U64,
	.desc	= "CPU time (system) per second of childs",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_csttime_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 'c', "cste"),
	.offset = SD_TASK_OFFSET(cpu_csttime_diff),
	.type	= SD_TYPE_U64,
	.desc	= "CPU time (steal) per second of childs",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_cutime = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_us, 'R', "userC"),
	.offset = SD_TASK_OFFSET(ts.ac_cutime),
	.type	= SD_TYPE_U64,
	.desc	= "CPU time (user) of childs",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_cstime = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_us, 'Y', "sysC"),
	.offset = SD_TASK_OFFSET(ts.ac_cstime),
	.type	= SD_TYPE_U64,
	.desc	= "CPU time (system) of childs",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_csttime = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_us, 'A',
					     "stealC"),
	.offset = SD_TASK_OFFSET(ts.ac_csttime),
	.type	= SD_TYPE_U64,
	.desc	= "CPU time (steal) of childs",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_user = {
	.table_col = TABLE_COL_TIME_SUM(table_col_unit_us, 'U', "user+"),
	.offset = SD_TASK_OFFSET(ts.ac_utime),
	.type	= SD_TYPE_U64,
	.desc	= "Total CPU time (user)",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_sys = {
	.table_col = TABLE_COL_TIME_SUM(table_col_unit_us, 'S', "sys+"),
	.offset = SD_TASK_OFFSET(ts.ac_stime),
	.type	= SD_TYPE_U64,
	.desc	= "Total CPU time (system)",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_cpu_steal = {
	.table_col = TABLE_COL_TIME_SUM(table_col_unit_us, 'T', "ste+"),
	.offset = SD_TASK_OFFSET(ts.ac_sttime),
	.type	= SD_TYPE_U64,
	.desc	= "Total CPU time (steal)",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_etime = {
	.table_col = TABLE_COL_TIME_MAX(table_col_unit_hm, 'E', "elap+"),
	.offset = SD_TASK_OFFSET(ts.ac_etime),
	.type	= SD_TYPE_U64,
	.desc	= "Elapsed time",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_delay_cpu = {
	.table_col = TABLE_COL_TIME_MAX(table_col_unit_us, 'D', "delay+"),
	.offset = SD_TASK_OFFSET(ts.cpu_delay_total),
	.type	= SD_TYPE_U64,
	.desc	= "Total CPU delay time",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_delay_cpu_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 'd', "delay"),
	.offset = SD_TASK_OFFSET(delay_cpu_diff),
	.type	= SD_TYPE_U64,
	.desc	= "CPU delay time per second",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_delay_blkio_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 'b', "blkio"),
	.offset = SD_TASK_OFFSET(delay_blkio_diff),
	.type	= SD_TYPE_U64,
	.desc	= "Block IO delay time per second",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_delay_swap_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 'w', "swap"),
	.offset = SD_TASK_OFFSET(delay_swap_diff),
	.type	= SD_TYPE_U64,
	.desc	= "Swap delay time per second",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_delay_mem_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 'e', "memdel"),
	.offset = SD_TASK_OFFSET(delay_mem_diff),
	.type	= SD_TYPE_U64,
	.desc	= "Memory delay time per second",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_total_diff = {
	.table_col = TABLE_COL_TIME_DIFF_SUM(table_col_unit_perc, 'o', "total"),
	.offset = SD_TASK_OFFSET(total_diff),
	.type	= SD_TYPE_U64,
	.desc	= "Total used time per second",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_page_fault_min = {
	.table_col = TABLE_COL_CNT_SUM('p', "pfmin"),
	.offset = SD_TASK_OFFSET(ts.ac_minflt),
	.type	= SD_TYPE_U64,
	.desc	= "Minor page fault count",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_page_fault_maj = {
	.table_col = TABLE_COL_CNT_SUM('P', "pfmaj"),
	.offset = SD_TASK_OFFSET(ts.ac_majflt),
	.type	= SD_TYPE_U64,
	.desc	= "Major page fault count",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_csw_vol = {
	.table_col = TABLE_COL_CNT_SUM('v', "cswi_v"),
	.offset = SD_TASK_OFFSET(ts.nvcsw),
	.type	= SD_TYPE_U64,
	.desc	= "Voluntary context switches",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_csw_nonvol = {
	.table_col = TABLE_COL_CNT_SUM('n', "cswi_n"),
	.offset = SD_TASK_OFFSET(ts.nivcsw),
	.type	= SD_TYPE_U64,
	.desc	= "Nonvoluntary context switches",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_mem_core = {
	.table_col = TABLE_COL_MEM_SUM(table_col_unit_mib, 'm', "memcore"),
	.offset	= SD_TASK_OFFSET(mem_core_kib),
	.type	= SD_TYPE_U64,
	.desc	= "Core memory",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_mem_virt = {
	.table_col = TABLE_COL_MEM_SUM(table_col_unit_mib, 'M', "memvirt"),
	.offset	= SD_TASK_OFFSET(mem_virt_kib),
	.type	= SD_TYPE_U64,
	.desc	= "Virtual memory",
	.fn_u64	= l_task_item_u64,
};

struct sd_task_item sd_task_item_name = {
	.table_col = TABLE_COL_STR_LEFT('N', "name"),
	.offset	= SD_TASK_OFFSET(name.process),
	.type	= SD_TYPE_STR,
	.desc	= "Process name",
};
