/*
 * ptop - Precise top command using the taskstats interface
 *
 * Window "task_list":
 * Shows a list of tasks that the system is currently running.
 *
 * Copyright IBM Corp. 2010
 * Author(s): Michael Holzheu <holzheu@linux.vnet.ibm.com>
 */

#include "helper.h"
#include "table.h"
#include "ptop.h"
#include "win_fields.h"
#include "win_help.h"
#include "sd.h"
#include "nav_desc.h"
#include "opts.h"

/*
 * Globals for task_list window
 */
static struct table		*l_t;		/* Table */
static struct ptop_win	*l_win_fields;	/* Fields Window */
static struct ptop_win	*l_win_help;	/* Herp Window */

/* PID column */
static struct table_col l_col_pid = {
	.type		= TABLE_COL_TYPE_U64,
	.unit		= &table_col_unit_cnt,
	.unit_fam	= table_col_unit_fam_cnt,
	.align		= TABLE_COL_ALIGN_LEFT,
	.agg		= TABLE_COL_AGG_NONE,
	.hotkey		= 'i',
	.head		= "pid  ", /* XXX: Keep trailing blanks */
};

/*
 * Online help text for task_list window
 */
static const char l_help_str[] =
"\\BNavigation\\B\n"
"\n"
"To navigate between the windows, use the arrow keys or 'hjkl'. The windows\n"
"have two modes, \"normal mode\" and \"select mode\". When you start the "
"program,\n"
"the window is in normal mode where data is updated at regular intervals. Use\n"
"the RIGHT arrow key to enter the select mode. In select mode you can select\n"
"rows with with UP and DOWN arrow keys and mark them with the SPACE bar. From\n"
"the \"task_list\" window you can access the \"task\" window in select mode\n"
"with the arrow key RIGHT. Leave the select mode with the arrow key LEFT.\n"
"If you are in normal mode, the arrow key LEFT goes to the previous window.\n"
"You can scroll all windows using the arrow keys UP, DOWN, PAGEUP and\n"
"PAGEDOWN. You can jump to the end of a window with 'G' and to the beginning\n"
"with 'g'.\n"
"\n"
"Select a column by pressing the hotkey of the column. This key is underlined\n"
"in the heading. The table is sorted according to the values in the selected\n"
"column. If you press the hotkey again, the sort order is reversed.\n"
"Alternatively you can select columns with the '<' and '>' keys.\n"
"\n"
"\\BTable layout\\B\n"
"\n"
"At the top left of the table the current time is shown. Then the CPU types\n"
"with the physical CPU numbers that are used for CPU time calculation are\n"
"displayed. The second row shows the units that are currently used for\n"
"formatting the data. The last row shows the status display (see description\n"
"below) and the aggregation of the the data columns. The last row aggregates\n"
"all rows, not only the visible ones. If only the marked rows are shown\n"
"(with '.') then only these rows are aggregated.\n"
"\n"
"\\BFields window\\B\n"
"\n"
"Press 'f' to enter the fields window. In this window you can select and\n"
"deselect fields and you can modify the units for the fields.\n"
"\n"
"\\BStatus display\\B\n"
"\n"
"At the left bottom of the screen a status display is shown.\n"
"Example: \"V:V:N\"\n\n"
"The first character shows, if the window can be scrolled:\n"
" 'V': Window can be scrolled down\n"
" '|': Window can be scrolled up/down\n"
" '^': Window can be scrolled up\n"
" '=': Window cannot be scrolled\n"
"The second character shows the sort order for sorted tables:\n"
" 'V': Higher values first\n"
" '^': Lower values first\n"
"The third character shows the current mode:\n"
" 'N': Normal mode\n"
" 'S': Select mode\n";

/*
 * Description of Navigation Keys (used for help window)
 */
static struct nav_desc *l_nav_desc_normal_vec[] = {
	&nav_desc_select_mode_enter,
	&nav_desc_marks_clear,
	NULL,
};

static struct nav_desc *l_nav_desc_select_vec[] = {
	&nav_desc_select_mode_leave,
	&nav_desc_mark_toggle,
	NULL,
};

static struct nav_desc *l_nav_desc_general_vec[] = {
	&nav_desc_win_enter_fields,
	&nav_desc_col_unit_increase,
	&nav_desc_col_unit_decrease,
	&nav_desc_select_col_next,
	&nav_desc_select_col_prev,
	&nav_desc_select_col_hotkey,
	&nav_desc_scroll_up_line,
	&nav_desc_scroll_down_line,
	&nav_desc_scroll_up_page,
	&nav_desc_scroll_down_page,
	&nav_desc_scroll_up_head,
	&nav_desc_scroll_down_tail,
	&nav_desc_mark_toggle_view,
	&nav_desc_quit,
	NULL,
};

/*
 * Add task item to table row
 */
static void l_task_item_add(struct table_row *table_row, struct sd_task *task,
			   struct sd_task_item *item)
{
	switch (sd_task_item_type(item)) {
	case SD_TYPE_U64:
	case SD_TYPE_U32:
	case SD_TYPE_U16:
		table_row_entry_u64_add(table_row,
					sd_task_item_table_col(item),
					sd_task_item_u64(task, item));
		break;
	case SD_TYPE_S64:
		table_row_entry_s64_add(table_row,
					sd_task_item_table_col(item),
					sd_task_item_s64(task, item));
		break;
	case SD_TYPE_STR:
		table_row_entry_str_add(table_row,
					sd_task_item_table_col(item),
					sd_task_item_str(task, item));
		break;
	}
}

/*
 * Add tasktem to table
 */
static void l_task_add(struct sd_task *task)
{
	struct table_row *table_row;
	struct sd_task_item *item;
	unsigned int i;

	table_row = table_row_alloc(l_t);
	// table_row_entry_str_add(table_row, &l_col_pid, sd_task_id(task));
	table_row_entry_u64_add(table_row, &l_col_pid, atoi(sd_task_id(task)));

	sd_task_item_iterate(item, i)
		l_task_item_add(table_row, task, item);
	table_row_add(l_t, table_row);
}

/*
 * Fill tasktem data into table
 */
static void l_table_create(void)
{
	struct sd_task *parent, *guest;

	table_row_del_all(l_t);
	parent = sd_task_root_get();
	sd_task_iterate(parent, guest) {
		if (guest->i.calculated < 2)
			continue; // XXX
		if (!opts_task_specified(&win_task_list, sd_task_id(guest)))
			continue;
		l_task_add(guest);
	}
	table_finish(l_t);
}

/*
 * Print first line
 */
void l_print_head(void)
{
	ht_print_time();
	ptop_printf(" ");
	ht_print_help_icon();
	ptop_print_nl();
}

/*
 * Print table to screen
 */
static void l_table_update_term(struct ptop_win *win)
{
	(void) win;

	l_print_head();
	table_print(l_t);
}

/*
 * Process input and switch window if necessary
 */
static enum ptop_win_action l_process_input(struct ptop_win *win, int c)
{
	// char selected_task[TABLE_STR_MAX];
	(void) win;

	switch (c) {
	case 'f':
		return win_switch(l_win_fields);
	case '?':
		return win_switch(l_win_help);
	case 'q':
		ptop_exit(0);
#if 0
	case 'l':
	case KEY_RIGHT:
		if (!table_mode_select(l_t))
			break;
		table_row_select_key_get(l_t, selected_task);
		win_task_set(selected_task);
		return win_switch(&win_task);
#endif
	case ERR:
		break;
	}
	table_process_input(l_t, c);
	ptop_update_term();
	return WIN_KEEP;
}

/*
 * Enable field and set unit
 */
static void l_field_set(struct table_col_spec *col_spec)
{
	table_col_enable_toggle(l_t, col_spec->hotkey);
	if (!col_spec->unit_str)
		return;
	if (table_col_unit_set(l_t, col_spec->hotkey, col_spec->unit_str))
		ERR_EXIT("Invalid unit \"%s\" for field \"%c\"\n",
			 col_spec->unit_str, col_spec->hotkey);
}

/*
 * Enable field defined in "col_spec"
 */
static void l_field_enable(struct table_col_spec *col_spec)
{
	struct sd_task_item *item;
	struct table_col *col;
	unsigned int i;

	sd_task_item_iterate(item, i) {
		col = sd_task_item_table_col(item);
		if (table_col_hotkey(col) != col_spec->hotkey)
			continue;
		l_field_set(col_spec);
		return;
	}
	ERR_EXIT("Unknown field \"%c\"\n", col_spec->hotkey);
}

/*
 * Enable fields defined on command line
 */
static void l_fields_enable_cmdline(void)
{
	unsigned int i;

	for (i = 0; i < win_task_list.opts.fields.cnt; i++)
		l_field_enable(win_task_list.opts.fields.vec[i]);
}

/*
 * Enable fields like defined in data gatherer
 */
static void l_fields_enable_default(void)
{
	struct sd_task_item *item;
	struct table_col *col;
	unsigned int i;

	sd_task_item_enable_iterate(item, i) {
		col = sd_task_item_table_col(item);
		table_col_enable_toggle(l_t, table_col_hotkey(col));
	}
}

/*
 * Event loop: Make regular updates of table
 */
static void l_run(struct ptop_win *win)
{
	enum ptop_win_action action;
	(void) win;

	/* Reformat table when entering window */
	table_rebuild(l_t);
	while (1) {
		l_table_create();
		ptop_update_term();
		action = ptop_process_input_timeout();
		if (action == WIN_SWITCH)
			return;
		/* No updates in select mode */
		if (!table_mode_select(l_t))
			sd_update();
	}
}

/*
 * Initialize window
 */
void win_task_list_init(void)
{
	struct table_col **col_vec;
	struct sd_task_item *item;
	struct table_col *col;
	char **col_desc_vec;
	unsigned int i;
	int item_cnt;

	/* Alloc table and add columns */
	l_t = table_new(1, 1, 1, 1);
	table_col_add(l_t, &l_col_pid);

	item_cnt = sd_task_item_cnt() + 1;
	col_vec = ht_zalloc(sizeof(void *) * item_cnt);
	col_desc_vec = ht_zalloc(sizeof(void *) * item_cnt);

	sd_task_item_iterate(item, i) {
		col = sd_task_item_table_col(item);
		table_col_add(l_t, col);
		table_col_enable_toggle(l_t, table_col_hotkey(col));
		col_vec[i] = col;
		col_desc_vec[i] = item->desc;
	}
	/* Enable fields */
	if (win_task_list.opts.fields.specified)
		l_fields_enable_cmdline();
	else
		l_fields_enable_default();

	/* Select sort field */
	if (win_task_list.opts.sort_field_specified) {
		for (i = 0; i < win_task_list.opts.sort_field_specified; i++) {
			if (table_col_select(l_t, win_task_list.opts.sort_field))
				ERR_EXIT("Sort field \"%c\" is not available\n",
					  win_task_list.opts.sort_field);
		}
	} else {
		table_col_select(l_t, sd_task_item_total_diff.table_col.hotkey);
	}
	/* Initialize help and fields window */
	l_win_help = win_help_new(&win_task_list);
	l_win_fields = win_fields_new(l_t, col_vec, col_desc_vec);
}

/*
 * ptop window structure definition
 */
struct ptop_win win_task_list = {
	.process_input		= l_process_input,
	.update_term		= l_table_update_term,
	.run			= l_run,
	.id			= "task_list",
	.desc			= l_help_str,
	.desc_normal_vec	= l_nav_desc_normal_vec,
	.desc_select_vec	= l_nav_desc_select_vec,
	.desc_general_vec	= l_nav_desc_general_vec,
};
